"""Keyboard handling mixin for LikX EditorWindow."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import gi

gi.require_version("Gdk", "3.0")
gi.require_version("Gtk", "3.0")
from gi.repository import Gdk, Gtk  # noqa: E402

from ..capture import CaptureResult  # noqa: E402
from ..editor import ToolType  # noqa: E402

if TYPE_CHECKING:
    pass


class KeyboardMixin:
    """Mixin providing keyboard handling for EditorWindow.

    This mixin expects the following attributes on the class:
    - editor_state: EditorState
    - drawing_area: Gtk.DrawingArea
    - statusbar: Gtk.Statusbar
    - statusbar_context: int
    - notebook: Gtk.Notebook
    - tabs: List[TabContent]
    - current_tab_index: int
    - tool_buttons: Dict[ToolType, Gtk.ToggleButton]
    - save_handler: SaveHandler

    And the following methods:
    - _show_toast(message: str)
    - _set_tool(tool: ToolType)
    - _show_command_palette()
    - _update_zoom_label()
    - close_tab(index: int)
    """

    def _init_key_bindings(self) -> None:
        """Initialize keyboard shortcut dispatch tables."""
        # Tool shortcuts (no modifiers)
        self._tool_shortcuts = {
            Gdk.KEY_p: ToolType.PEN,
            Gdk.KEY_h: ToolType.HIGHLIGHTER,
            Gdk.KEY_l: ToolType.LINE,
            Gdk.KEY_a: ToolType.ARROW,
            Gdk.KEY_r: ToolType.RECTANGLE,
            Gdk.KEY_e: ToolType.ELLIPSE,
            Gdk.KEY_t: ToolType.TEXT,
            Gdk.KEY_b: ToolType.BLUR,
            Gdk.KEY_x: ToolType.PIXELATE,
            Gdk.KEY_m: ToolType.MEASURE,
            Gdk.KEY_n: ToolType.NUMBER,
            Gdk.KEY_i: ToolType.COLORPICKER,
            Gdk.KEY_s: ToolType.STAMP,
            Gdk.KEY_z: ToolType.ZOOM,
            Gdk.KEY_k: ToolType.CALLOUT,
            Gdk.KEY_c: ToolType.CROP,
            Gdk.KEY_v: ToolType.SELECT,
        }

        # Ctrl+Alt alignment shortcuts: (key, method, success_msg)
        self._align_shortcuts = {
            Gdk.KEY_l: ("align_left", "Aligned left"),
            Gdk.KEY_L: ("align_left", "Aligned left"),
            Gdk.KEY_r: ("align_right", "Aligned right"),
            Gdk.KEY_R: ("align_right", "Aligned right"),
            Gdk.KEY_t: ("align_top", "Aligned top"),
            Gdk.KEY_T: ("align_top", "Aligned top"),
            Gdk.KEY_b: ("align_bottom", "Aligned bottom"),
            Gdk.KEY_B: ("align_bottom", "Aligned bottom"),
            Gdk.KEY_c: ("align_center_horizontal", "Aligned center (horizontal)"),
            Gdk.KEY_C: ("align_center_horizontal", "Aligned center (horizontal)"),
            Gdk.KEY_m: ("align_center_vertical", "Aligned center (vertical)"),
            Gdk.KEY_M: ("align_center_vertical", "Aligned center (vertical)"),
            Gdk.KEY_w: ("match_width", "Matched width"),
            Gdk.KEY_W: ("match_width", "Matched width"),
            Gdk.KEY_e: ("match_height", "Matched height"),
            Gdk.KEY_E: ("match_height", "Matched height"),
            Gdk.KEY_s: ("match_size", "Matched size"),
            Gdk.KEY_S: ("match_size", "Matched size"),
            Gdk.KEY_f: ("flip_vertical", "Flipped vertically"),
            Gdk.KEY_F: ("flip_vertical", "Flipped vertically"),
        }

    def _handle_alignment_shortcut(self, key: int) -> bool:
        """Handle Ctrl+Alt alignment shortcuts."""
        if key not in self._align_shortcuts:
            return False
        method_name, success_msg = self._align_shortcuts[key]
        method = getattr(self.editor_state, method_name)
        if method():
            self._show_toast(success_msg)
            self.drawing_area.queue_draw()
        else:
            self._show_toast("Select 2+ elements")
        return True

    def _ctrl_copy(self) -> None:
        """Handle Ctrl+C (copy)."""
        if self.editor_state.selected_indices:
            if self.editor_state.copy_selected():
                self._show_toast(f"Copied {len(self.editor_state.selected_indices)} annotation(s)")
        else:
            self.save_handler.copy_to_clipboard()

    def _ctrl_paste(self) -> None:
        """Handle Ctrl+V (paste annotations, or image from system clipboard as new tab)."""
        if self.editor_state.has_clipboard() and self.editor_state.paste_annotations():
            self._show_toast(f"Pasted {len(self.editor_state.selected_indices)} annotation(s)")
            self.drawing_area.queue_draw()
            return

        # No annotation clipboard — try pasting image from system clipboard
        self._paste_image_from_clipboard()

    def _ctrl_lock_toggle(self) -> None:
        """Handle Ctrl+L (lock toggle)."""
        if self.editor_state.toggle_lock_selected():
            state = "Locked" if self.editor_state.is_selection_locked() else "Unlocked"
            self._show_toast(state)
            self.drawing_area.queue_draw()
        else:
            self._show_toast("Select element(s) to lock")

    def _ctrl_grid_toggle(self) -> None:
        """Handle Ctrl+' (grid snap toggle)."""
        new_state = not self.editor_state.grid_snap_enabled
        self.editor_state.set_grid_snap(new_state)
        self._show_toast("Grid snap ON" if new_state else "Grid snap OFF")
        self.drawing_area.queue_draw()

    def _handle_ctrl_shortcuts(self, key: int, shift: bool) -> bool:
        """Handle Ctrl+key shortcuts. Returns True if handled."""
        # Simple shortcuts with immediate actions
        simple_shortcuts = {
            Gdk.KEY_s: self.save_handler.save,
            Gdk.KEY_z: self._undo,
            Gdk.KEY_y: self._redo,
            Gdk.KEY_c: self._ctrl_copy,
            Gdk.KEY_v: self._ctrl_paste,
            Gdk.KEY_l: self._ctrl_lock_toggle,
            Gdk.KEY_apostrophe: self._ctrl_grid_toggle,
        }

        if key in simple_shortcuts:
            simple_shortcuts[key]()
            return True

        # Shortcuts that need success check and toast
        action_shortcuts = {
            Gdk.KEY_bracketright: (self.editor_state.bring_to_front, "Brought to front"),
            Gdk.KEY_bracketleft: (self.editor_state.send_to_back, "Sent to back"),
            Gdk.KEY_d: (
                self.editor_state.duplicate_selected,
                lambda: f"Duplicated {len(self.editor_state.selected_indices)} annotation(s)",
            ),
        }

        if key in action_shortcuts:
            action, msg = action_shortcuts[key]
            if action():
                self._show_toast(msg() if callable(msg) else msg)
                self.drawing_area.queue_draw()
            return True

        return False

    def _handle_ctrl_shift_shortcuts(self, key: int) -> bool:
        """Handle Ctrl+Shift+key shortcuts. Returns True if handled."""
        if key in (Gdk.KEY_p, Gdk.KEY_P):
            self._show_command_palette()
            return True

        # Shortcuts with action/success/fail pattern
        shortcuts = {
            (Gdk.KEY_h, Gdk.KEY_H): (
                self.editor_state.distribute_horizontal,
                "Distributed horizontally",
                "Select 3+ elements to distribute",
            ),
            (Gdk.KEY_j, Gdk.KEY_J): (
                self.editor_state.distribute_vertical,
                "Distributed vertically",
                "Select 3+ elements to distribute",
            ),
            (Gdk.KEY_g, Gdk.KEY_G): (
                self.editor_state.ungroup_selected,
                "Ungrouped",
                "No groups to ungroup",
            ),
            (Gdk.KEY_f, Gdk.KEY_F): (
                self.editor_state.flip_horizontal,
                "Flipped horizontally",
                "Select element(s) to flip",
            ),
            (Gdk.KEY_r, Gdk.KEY_R): (
                lambda: self.editor_state.rotate_selected(-90),
                "Rotated -90°",
                "Select element(s) to rotate",
            ),
        }

        for keys, (action, success_msg, fail_msg) in shortcuts.items():
            if key in keys:
                if action():
                    self._show_toast(success_msg)
                    self.drawing_area.queue_draw()
                else:
                    self._show_toast(fail_msg)
                return True

        return False

    def _handle_tab_switch(self, key: int, shift: bool) -> bool:
        """Handle Ctrl+Tab / Ctrl+Shift+Tab for tab switching."""
        if key != Gdk.KEY_Tab or len(self.tabs) <= 1:
            return key == Gdk.KEY_Tab
        offset = -1 if shift else 1
        self.notebook.set_current_page((self.current_tab_index + offset) % len(self.tabs))
        return True

    def _handle_opacity_adjust(self, key: int) -> bool:
        """Handle Shift+[ / Shift+] for opacity adjustment."""
        adjustments = {Gdk.KEY_bracketleft: -0.1, Gdk.KEY_bracketright: 0.1}
        if key not in adjustments:
            return False
        if self.editor_state.adjust_selected_opacity(adjustments[key]):
            opacity = self.editor_state.get_selected_opacity() or 1.0
            self._show_toast(f"Opacity: {int(opacity * 100)}%")
            self.drawing_area.queue_draw()
        else:
            self._show_toast("Select element(s) to adjust opacity")
        return True

    def _handle_zoom_shortcuts(self, key: int) -> bool:
        """Handle zoom shortcuts (+/-/0)."""
        if key in (Gdk.KEY_plus, Gdk.KEY_equal, Gdk.KEY_KP_Add):
            self.editor_state.zoom_in()
        elif key in (Gdk.KEY_minus, Gdk.KEY_KP_Subtract):
            self.editor_state.zoom_out()
        elif key == Gdk.KEY_0:
            self.editor_state.reset_zoom()
        else:
            return False
        self._update_zoom_label()
        self.drawing_area.queue_draw()
        return True

    def _handle_arrow_nudge(self, key: int, shift: bool) -> bool:
        """Handle arrow key nudging."""
        nudge = 10 if shift else 1
        offsets = {
            Gdk.KEY_Up: (0, -nudge),
            Gdk.KEY_Down: (0, nudge),
            Gdk.KEY_Left: (-nudge, 0),
            Gdk.KEY_Right: (nudge, 0),
        }
        if key not in offsets:
            return False
        if self.editor_state.nudge_selected(*offsets[key]):
            self.drawing_area.queue_draw()
        return True

    def _handle_ctrl_key(self, key: int, shift: bool, alt: bool) -> bool:
        """Handle Ctrl+key combinations."""
        if self._handle_tab_switch(key, shift):
            return True
        if not shift and key in (Gdk.KEY_w, Gdk.KEY_W):
            self.close_tab(self.current_tab_index)
            return True
        if not shift and key in (Gdk.KEY_g, Gdk.KEY_G):
            if self.editor_state.group_selected():
                self._show_toast(f"Grouped {len(self.editor_state.selected_indices)} elements")
                self.drawing_area.queue_draw()
            else:
                self._show_toast("Select 2+ elements to group")
            return True
        if not shift and not alt and key in (Gdk.KEY_r, Gdk.KEY_R):
            if self.editor_state.rotate_selected(90):
                self._show_toast("Rotated 90°")
                self.drawing_area.queue_draw()
            else:
                self._show_toast("Select element(s) to rotate")
            return True
        if shift and self._handle_ctrl_shift_shortcuts(key):
            return True
        if alt and self._handle_alignment_shortcut(key):
            return True
        if not shift and not alt and self._handle_ctrl_shortcuts(key, shift):
            return True
        return False

    def _handle_tool_shortcut(self, key: int) -> bool:
        """Handle tool selection shortcuts."""
        if key not in self._tool_shortcuts:
            return False
        tool = self._tool_shortcuts[key]
        self._set_tool(tool)
        if hasattr(self, "tool_buttons") and tool in self.tool_buttons:
            self.tool_buttons[tool].set_active(True)
        return True

    def _handle_escape(self) -> bool:
        """Handle Escape key."""
        if self.editor_state.selected_index is not None:
            self.editor_state.deselect()
        else:
            self.editor_state.cancel_drawing()
        self.drawing_area.queue_draw()
        return True

    def _on_key_press(self, widget: Gtk.Widget, event: Gdk.EventKey) -> bool:
        """Handle keyboard shortcuts using dispatch tables."""
        ctrl = event.state & Gdk.ModifierType.CONTROL_MASK
        shift = event.state & Gdk.ModifierType.SHIFT_MASK
        alt = event.state & Gdk.ModifierType.MOD1_MASK
        key = event.keyval

        if ctrl and self._handle_ctrl_key(key, shift, alt):
            return True

        if shift and not ctrl and not alt and self._handle_opacity_adjust(key):
            return True

        if self._handle_tool_shortcut(key):
            return True

        if self._handle_zoom_shortcuts(key):
            return True

        if key in (Gdk.KEY_Delete, Gdk.KEY_BackSpace):
            if self.editor_state.delete_selected():
                self.statusbar.push(self.statusbar_context, "Element deleted")
                self.drawing_area.queue_draw()
            return True

        if self._handle_arrow_nudge(key, shift):
            return True

        if key == Gdk.KEY_Escape:
            return self._handle_escape()

        return False

    def _paste_image_from_clipboard(self) -> None:
        """Paste an image from the system clipboard as a new tab."""
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        pixbuf = clipboard.wait_for_image()

        if pixbuf is None:
            self._show_toast("No image on clipboard")
            return

        result = CaptureResult(success=True, pixbuf=pixbuf)

        if hasattr(self, "add_tab"):
            self.add_tab(result, switch_to=True)
            # Show tab bar now that we have multiple tabs
            if hasattr(self, "notebook"):
                self.notebook.set_show_tabs(True)
            self._show_toast("Pasted image from clipboard")
        else:
            logging.warning("Cannot paste image: add_tab not available")



"""Screenshot capture module for LikX with Wayland and X11 support."""

import os
import subprocess
import tempfile
import time
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import List, Optional, Tuple

try:
    import gi

    gi.require_version("Gdk", "3.0")
    gi.require_version("GdkPixbuf", "2.0")
    from gi.repository import Gdk, GdkPixbuf

    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

from . import config
from .clipboard import copy_to_clipboard  # noqa: F401


class CaptureMode(Enum):
    """Screenshot capture modes."""

    FULLSCREEN = "fullscreen"
    REGION = "region"
    WINDOW = "window"


class DisplayServer(Enum):
    """Display server types."""

    X11 = "x11"
    WAYLAND = "wayland"
    UNKNOWN = "unknown"


class CaptureResult:
    """Result of a screenshot capture operation."""

    def __init__(
        self,
        success: bool,
        filepath: Optional[Path] = None,
        error: Optional[str] = None,
        pixbuf: Optional[object] = None,
    ):
        self.success = success
        self.filepath = filepath
        self.error = error
        self.pixbuf = pixbuf

    def __bool__(self) -> bool:
        return self.success


@dataclass
class MonitorInfo:
    """Information about a display monitor."""

    index: int
    name: str
    x: int
    y: int
    width: int
    height: int
    is_primary: bool = False
    scale_factor: int = 1

    @property
    def geometry(self) -> Tuple[int, int, int, int]:
        """Return (x, y, width, height) tuple."""
        return (self.x, self.y, self.width, self.height)

    def __str__(self) -> str:
        primary = " (Primary)" if self.is_primary else ""
        return f"{self.name}: {self.width}x{self.height}{primary}"


def detect_display_server() -> DisplayServer:
    """Detect the current display server (X11 or Wayland).

    Returns:
        DisplayServer enum indicating the detected server.
    """
    session_type = os.environ.get("XDG_SESSION_TYPE", "").lower()
    wayland_display = os.environ.get("WAYLAND_DISPLAY", "")

    if "wayland" in session_type or wayland_display:
        return DisplayServer.WAYLAND
    elif os.environ.get("DISPLAY"):
        return DisplayServer.X11
    return DisplayServer.UNKNOWN


def get_monitors() -> List[MonitorInfo]:
    """Get information about all connected monitors.

    Returns:
        List of MonitorInfo objects for each monitor.
    """
    if not GTK_AVAILABLE:
        return []

    monitors = []
    display = Gdk.Display.get_default()
    if display is None:
        return []

    n_monitors = display.get_n_monitors()
    primary_monitor = display.get_primary_monitor()

    for i in range(n_monitors):
        monitor = display.get_monitor(i)
        if monitor is None:
            continue

        geometry = monitor.get_geometry()
        is_primary = monitor == primary_monitor

        # Get monitor name/model
        name = monitor.get_model() or f"Monitor {i + 1}"

        monitors.append(
            MonitorInfo(
                index=i,
                name=name,
                x=geometry.x,
                y=geometry.y,
                width=geometry.width,
                height=geometry.height,
                is_primary=is_primary,
                scale_factor=monitor.get_scale_factor(),
            )
        )

    return monitors


def get_monitor_at_point(x: int, y: int) -> Optional[MonitorInfo]:
    """Get the monitor containing a specific point.

    Args:
        x, y: Screen coordinates

    Returns:
        MonitorInfo for the monitor at that point, or None if not found.
    """
    for monitor in get_monitors():
        if (
            monitor.x <= x < monitor.x + monitor.width
            and monitor.y <= y < monitor.y + monitor.height
        ):
            return monitor
    return None


def get_primary_monitor() -> Optional[MonitorInfo]:
    """Get the primary monitor.

    Returns:
        MonitorInfo for the primary monitor, or None if not found.
    """
    monitors = get_monitors()
    for monitor in monitors:
        if monitor.is_primary:
            return monitor
    # Fallback to first monitor if no primary
    return monitors[0] if monitors else None


def capture_monitor(
    monitor: MonitorInfo, delay: int = 0, include_cursor: bool = False
) -> CaptureResult:
    """Capture a specific monitor.

    Args:
        monitor: MonitorInfo object for the monitor to capture.
        delay: Delay in seconds before capturing.
        include_cursor: Whether to include the mouse cursor.

    Returns:
        CaptureResult with the captured screenshot.
    """
    if delay > 0:
        time.sleep(delay)

    # Use region capture with monitor geometry
    return capture_region(
        *monitor.geometry,
        delay=0,  # Already delayed above
    )


def capture_fullscreen_wayland(delay: int = 0) -> CaptureResult:
    """Capture fullscreen on Wayland using available tools.

    Args:
        delay: Delay in seconds before capturing.

    Returns:
        CaptureResult with the captured screenshot.
    """
    if delay > 0:
        time.sleep(delay)

    fd, temp_file = tempfile.mkstemp(suffix=".png", prefix="likx_")
    os.close(fd)

    try:
        # Try grim (wlroots compositors like Sway)
        try:
            result = subprocess.run(["grim", temp_file], capture_output=True, timeout=5)
            if result.returncode == 0 and os.path.exists(temp_file):
                pixbuf = GdkPixbuf.Pixbuf.new_from_file(temp_file)
                if pixbuf is not None:
                    return CaptureResult(True, pixbuf=pixbuf)
        except (subprocess.TimeoutExpired, FileNotFoundError, Exception):
            pass

        # Try gnome-screenshot
        try:
            result = subprocess.run(
                ["gnome-screenshot", "-f", temp_file], capture_output=True, timeout=5
            )
            if result.returncode == 0 and os.path.exists(temp_file):
                pixbuf = GdkPixbuf.Pixbuf.new_from_file(temp_file)
                if pixbuf is not None:
                    return CaptureResult(True, pixbuf=pixbuf)
        except (subprocess.TimeoutExpired, FileNotFoundError, Exception):
            pass

        # Try spectacle (KDE)
        try:
            result = subprocess.run(
                ["spectacle", "-b", "-n", "-o", temp_file], capture_output=True, timeout=5
            )
            if result.returncode == 0 and os.path.exists(temp_file):
                pixbuf = GdkPixbuf.Pixbuf.new_from_file(temp_file)
                if pixbuf is not None:
                    return CaptureResult(True, pixbuf=pixbuf)
        except (subprocess.TimeoutExpired, FileNotFoundError, Exception):
            pass
    finally:
        # Always clean up temp file
        try:
            os.unlink(temp_file)
        except OSError:
            pass

    return CaptureResult(
        False,
        error="No Wayland screenshot tool found. Install grim, gnome-screenshot, or spectacle.",
    )


def capture_fullscreen(delay: int = 0, include_cursor: bool = False) -> CaptureResult:
    """Capture the entire screen.

    Args:
        delay: Delay in seconds before capturing.
        include_cursor: Whether to include the mouse cursor.

    Returns:
        CaptureResult with the captured screenshot.
    """
    if delay > 0:
        time.sleep(delay)

    # Check display server
    display_server = detect_display_server()

    if display_server == DisplayServer.WAYLAND:
        return capture_fullscreen_wayland(0)

    if not GTK_AVAILABLE:
        return CaptureResult(False, error="GTK not available")

    try:
        # Get the default screen
        screen = Gdk.Screen.get_default()
        if screen is None:
            return CaptureResult(False, error="Could not get default screen")

        # Get the root window
        root_window = screen.get_root_window()
        if root_window is None:
            return CaptureResult(False, error="Could not get root window")

        # Get screen dimensions
        width = screen.get_width()
        height = screen.get_height()

        # Capture the screenshot
        pixbuf = Gdk.pixbuf_get_from_window(root_window, 0, 0, width, height)

        if pixbuf is None:
            return CaptureResult(False, error="Failed to capture screenshot")

        return CaptureResult(True, pixbuf=pixbuf)

    except Exception as e:
        return CaptureResult(False, error=str(e))


def capture_region_wayland(
    x: int, y: int, width: int, height: int, delay: int = 0
) -> CaptureResult:
    """Capture a region on Wayland.

    Args:
        x: X coordinate of the region.
        y: Y coordinate of the region.
        width: Width of the region.
        height: Height of the region.
        delay: Delay in seconds before capturing.

    Returns:
        CaptureResult with the captured screenshot.
    """
    if delay > 0:
        time.sleep(delay)

    fd, temp_file = tempfile.mkstemp(suffix=".png", prefix="likx_")
    os.close(fd)

    try:
        # Try grim with geometry (works with slurp output)
        try:
            geometry = f"{x},{y} {width}x{height}"
            result = subprocess.run(
                ["grim", "-g", geometry, temp_file], capture_output=True, timeout=5
            )
            if result.returncode == 0 and os.path.exists(temp_file):
                pixbuf = GdkPixbuf.Pixbuf.new_from_file(temp_file)
                if pixbuf is not None:
                    return CaptureResult(True, pixbuf=pixbuf)
        except (subprocess.TimeoutExpired, FileNotFoundError, Exception):
            pass

        # Try gnome-screenshot with area (GNOME Wayland)
        try:
            result = subprocess.run(
                ["gnome-screenshot", "-a", "-f", temp_file], capture_output=True, timeout=30
            )
            if result.returncode == 0 and os.path.exists(temp_file):
                pixbuf = GdkPixbuf.Pixbuf.new_from_file(temp_file)
                if pixbuf is not None:
                    return CaptureResult(True, pixbuf=pixbuf)
        except (subprocess.TimeoutExpired, FileNotFoundError, Exception):
            pass

        # Fallback: capture full screen and crop
        full_result = capture_fullscreen_wayland(0)
        if full_result.success and full_result.pixbuf:
            try:
                cropped = GdkPixbuf.Pixbuf.new(GdkPixbuf.Colorspace.RGB, True, 8, width, height)
                full_result.pixbuf.copy_area(x, y, width, height, cropped, 0, 0)
                return CaptureResult(True, pixbuf=cropped)
            except Exception as e:
                return CaptureResult(False, error=f"Failed to crop: {str(e)}")
    finally:
        try:
            os.unlink(temp_file)
        except OSError:
            pass

    return CaptureResult(False, error="Region capture failed on Wayland")


def capture_region(x: int, y: int, width: int, height: int, delay: int = 0) -> CaptureResult:
    """Capture a specific region of the screen.

    Args:
        x: X coordinate of the region.
        y: Y coordinate of the region.
        width: Width of the region.
        height: Height of the region.
        delay: Delay in seconds before capturing.

    Returns:
        CaptureResult with the captured screenshot.
    """
    if delay > 0:
        time.sleep(delay)

    # Check display server
    display_server = detect_display_server()

    if display_server == DisplayServer.WAYLAND:
        return capture_region_wayland(x, y, width, height, 0)

    if not GTK_AVAILABLE:
        return CaptureResult(False, error="GTK not available")

    try:
        screen = Gdk.Screen.get_default()
        if screen is None:
            return CaptureResult(False, error="Could not get default screen")

        root_window = screen.get_root_window()
        if root_window is None:
            return CaptureResult(False, error="Could not get root window")

        # Ensure coordinates are within screen bounds
        screen_width = screen.get_width()
        screen_height = screen.get_height()

        x = max(0, min(x, screen_width - 1))
        y = max(0, min(y, screen_height - 1))
        width = min(width, screen_width - x)
        height = min(height, screen_height - y)

        if width <= 0 or height <= 0:
            return CaptureResult(False, error="Invalid region dimensions")

        pixbuf = Gdk.pixbuf_get_from_window(root_window, x, y, width, height)

        if pixbuf is None:
            return CaptureResult(False, error="Failed to capture region")

        return CaptureResult(True, pixbuf=pixbuf)

    except Exception as e:
        return CaptureResult(False, error=str(e))


def capture_window_wayland(delay: int = 0) -> CaptureResult:
    """Capture active window on Wayland.

    Args:
        delay: Delay in seconds before capturing.

    Returns:
        CaptureResult with the captured screenshot.
    """
    if delay > 0:
        time.sleep(delay)

    fd, temp_file = tempfile.mkstemp(suffix=".png", prefix="likx_")
    os.close(fd)

    try:
        # Try gnome-screenshot with window flag
        try:
            result = subprocess.run(
                ["gnome-screenshot", "-w", "-f", temp_file], capture_output=True, timeout=5
            )
            if result.returncode == 0 and os.path.exists(temp_file):
                pixbuf = GdkPixbuf.Pixbuf.new_from_file(temp_file)
                if pixbuf is not None:
                    return CaptureResult(True, pixbuf=pixbuf)
        except (subprocess.TimeoutExpired, FileNotFoundError, Exception):
            pass

        # Try spectacle window mode
        try:
            result = subprocess.run(
                ["spectacle", "-a", "-b", "-n", "-o", temp_file],
                capture_output=True,
                timeout=5,
            )
            if result.returncode == 0 and os.path.exists(temp_file):
                pixbuf = GdkPixbuf.Pixbuf.new_from_file(temp_file)
                if pixbuf is not None:
                    return CaptureResult(True, pixbuf=pixbuf)
        except (subprocess.TimeoutExpired, FileNotFoundError, Exception):
            pass
    finally:
        try:
            os.unlink(temp_file)
        except OSError:
            pass

    return CaptureResult(
        False,
        error="Window capture not supported on Wayland without gnome-screenshot or spectacle",
    )


def _get_active_window_id() -> Tuple[Optional[int], Optional[str]]:
    """Get active window ID using xdotool.

    Returns:
        Tuple of (window_id, error_message). One will be None.
    """
    result = subprocess.run(
        ["xdotool", "getactivewindow"],
        capture_output=True,
        text=True,
        timeout=2,
    )
    if result.returncode != 0:
        return None, "Could not get active window. Is xdotool installed?"
    try:
        return int(result.stdout.strip()), None
    except ValueError:
        return None, f"Invalid window ID from xdotool: {result.stdout.strip()}"


def _get_window_geometry(window_id: int) -> Tuple[Optional[dict], Optional[str]]:
    """Get window geometry using xdotool.

    Returns:
        Tuple of (geometry_dict, error_message). One will be None.
    """
    result = subprocess.run(
        ["xdotool", "getwindowgeometry", "--shell", str(window_id)],
        capture_output=True,
        text=True,
        timeout=2,
    )
    if result.returncode != 0:
        return None, "Could not get window geometry"

    geometry = {}
    for line in result.stdout.split("\n"):
        if "=" in line:
            key, value = line.split("=", 1)
            try:
                geometry[key.strip()] = int(value.strip())
            except ValueError:
                pass

    required_keys = {"X", "Y", "WIDTH", "HEIGHT"}
    if not required_keys.issubset(geometry.keys()):
        missing = required_keys - geometry.keys()
        return None, f"Invalid geometry data, missing: {missing}"

    return geometry, None


def capture_window(window_id: Optional[int] = None, delay: int = 0) -> CaptureResult:
    """Capture a specific window.

    Args:
        window_id: X11 window ID (if None, captures active window).
        delay: Delay in seconds before capturing.

    Returns:
        CaptureResult with the captured screenshot.
    """
    if delay > 0:
        time.sleep(delay)

    display_server = detect_display_server()
    if display_server == DisplayServer.WAYLAND:
        return capture_window_wayland(0)

    if not GTK_AVAILABLE:
        return CaptureResult(False, error="GTK not available")

    try:
        if window_id is None:
            window_id, error = _get_active_window_id()
            if error:
                return CaptureResult(False, error=error)

        geometry, error = _get_window_geometry(window_id)
        if error:
            return CaptureResult(False, error=error)

        return capture_region(
            geometry["X"], geometry["Y"], geometry["WIDTH"], geometry["HEIGHT"], delay=0
        )

    except FileNotFoundError:
        return CaptureResult(
            False,
            error="xdotool not installed. Install it with: sudo apt install xdotool",
        )
    except subprocess.TimeoutExpired:
        return CaptureResult(False, error="Window capture timed out")
    except Exception as e:
        return CaptureResult(False, error=f"Window capture failed: {str(e)}")


def save_capture(
    result: CaptureResult,
    filepath: Optional[Path] = None,
    format_str: Optional[str] = None,
) -> CaptureResult:
    """Save a captured screenshot to file.

    Args:
        result: The CaptureResult containing the pixbuf.
        filepath: Path to save the file (auto-generated if None).
        format_str: Image format (uses config default if None).

    Returns:
        CaptureResult with the filepath set.
    """
    if not result.success or result.pixbuf is None:
        return CaptureResult(False, error="No screenshot to save")

    if filepath is None:
        filepath = config.get_save_path(format_str=format_str)
    else:
        filepath = Path(filepath)

    if format_str is None:
        format_str = filepath.suffix.lstrip(".").lower()
        if not format_str:
            format_str = config.get_setting("default_format", "png")

    # Map format strings to GdkPixbuf format names
    format_map = {
        "jpg": "jpeg",
        "jpeg": "jpeg",
        "png": "png",
        "bmp": "bmp",
        "gif": "gif",
    }

    pixbuf_format = format_map.get(format_str.lower(), "png")

    try:
        # Ensure parent directory exists
        filepath.parent.mkdir(parents=True, exist_ok=True)

        result.pixbuf.savev(str(filepath), pixbuf_format, [], [])
        return CaptureResult(True, filepath=filepath, pixbuf=result.pixbuf)

    except Exception as e:
        return CaptureResult(False, error=f"Failed to save: {str(e)}")


def capture(
    mode: CaptureMode,
    delay: int = 0,
    region: Optional[Tuple[int, int, int, int]] = None,
    window_id: Optional[int] = None,
    auto_save: bool = False,
    copy_clipboard: bool = True,
) -> CaptureResult:
    """Main capture function that handles all capture modes.

    Args:
        mode: The capture mode (fullscreen, region, window).
        delay: Delay in seconds before capturing.
        region: Tuple of (x, y, width, height) for region capture.
        window_id: Window ID for window capture.
        auto_save: Whether to automatically save the screenshot.
        copy_clipboard: Whether to copy to clipboard.

    Returns:
        CaptureResult with the captured screenshot.
    """
    cfg = config.load_config()

    # Use config values if not specified
    if delay == 0:
        delay = cfg.get("delay_seconds", 0)

    include_cursor = cfg.get("include_cursor", False)

    # Capture based on mode
    if mode == CaptureMode.FULLSCREEN:
        result = capture_fullscreen(delay=delay, include_cursor=include_cursor)
    elif mode == CaptureMode.REGION:
        if region is None:
            return CaptureResult(False, error="Region not specified")
        x, y, width, height = region
        result = capture_region(x, y, width, height, delay=delay)
    elif mode == CaptureMode.WINDOW:
        result = capture_window(window_id=window_id, delay=delay)
    else:
        return CaptureResult(False, error="Unknown capture mode")

    if not result.success:
        return result

    # Copy to clipboard if requested
    if copy_clipboard or cfg.get("copy_to_clipboard", True):
        copy_to_clipboard(result)

    # Auto-save if requested
    if auto_save or cfg.get("auto_save", False):
        result = save_capture(result)

    return result



"""Tests for the capture queue module."""

import tempfile
from datetime import datetime
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest


class TestQueueModuleImport:
    """Test queue module imports."""

    def test_module_imports(self):
        """Test that queue module imports successfully."""
        from src import queue

        assert hasattr(queue, "QueuedCapture")
        assert hasattr(queue, "CaptureQueue")
        assert hasattr(queue, "GTK_AVAILABLE")

    def test_gtk_available_is_bool(self):
        """Test that GTK_AVAILABLE is a boolean."""
        from src.queue import GTK_AVAILABLE

        assert isinstance(GTK_AVAILABLE, bool)


class TestQueuedCapture:
    """Test QueuedCapture dataclass."""

    def test_queued_capture_init(self):
        """Test QueuedCapture initialization."""
        from src.capture import CaptureMode
        from src.queue import QueuedCapture

        mock_result = MagicMock()
        timestamp = datetime.now()

        queued = QueuedCapture(
            result=mock_result,
            timestamp=timestamp,
            mode=CaptureMode.REGION,
        )

        assert queued.result == mock_result
        assert queued.timestamp == timestamp
        assert queued.mode == CaptureMode.REGION
        assert queued.temp_path is None

    def test_queued_capture_with_temp_path(self):
        """Test QueuedCapture with temp_path."""
        from src.capture import CaptureMode
        from src.queue import QueuedCapture

        queued = QueuedCapture(
            result=MagicMock(),
            timestamp=datetime.now(),
            mode=CaptureMode.FULLSCREEN,
            temp_path=Path("/tmp/test.png"),
        )

        assert queued.temp_path == Path("/tmp/test.png")


class TestCaptureQueueClass:
    """Test CaptureQueue class structure."""

    def test_class_has_add_method(self):
        """Test CaptureQueue has add method."""
        from src.queue import CaptureQueue

        assert hasattr(CaptureQueue, "add")
        assert callable(CaptureQueue.add)

    def test_class_has_get_all_method(self):
        """Test CaptureQueue has get_all method."""
        from src.queue import CaptureQueue

        assert hasattr(CaptureQueue, "get_all")

    def test_class_has_pop_all_method(self):
        """Test CaptureQueue has pop_all method."""
        from src.queue import CaptureQueue

        assert hasattr(CaptureQueue, "pop_all")

    def test_class_has_get_at_method(self):
        """Test CaptureQueue has get_at method."""
        from src.queue import CaptureQueue

        assert hasattr(CaptureQueue, "get_at")

    def test_class_has_clear_method(self):
        """Test CaptureQueue has clear method."""
        from src.queue import CaptureQueue

        assert hasattr(CaptureQueue, "clear")

    def test_class_has_remove_method(self):
        """Test CaptureQueue has remove method."""
        from src.queue import CaptureQueue

        assert hasattr(CaptureQueue, "remove")

    def test_class_has_count_property(self):
        """Test CaptureQueue has count property."""
        from src.queue import CaptureQueue

        assert hasattr(CaptureQueue, "count")

    def test_class_has_is_empty_property(self):
        """Test CaptureQueue has is_empty property."""
        from src.queue import CaptureQueue

        assert hasattr(CaptureQueue, "is_empty")


class TestCaptureQueueOperations:
    """Test CaptureQueue operations."""

    def test_queue_init_empty(self):
        """Test CaptureQueue initializes empty."""
        from src.queue import CaptureQueue

        with patch("src.queue.config.get_setting", return_value=50):
            queue = CaptureQueue()

            assert queue.count == 0
            assert queue.is_empty is True

    def test_queue_add_capture(self):
        """Test adding capture to queue."""
        from src.capture import CaptureMode
        from src.queue import CaptureQueue

        with patch("src.queue.config.get_setting", return_value=50):
            queue = CaptureQueue()

            mock_result = MagicMock()
            mock_result.success = True

            position = queue.add(mock_result, CaptureMode.REGION)

            assert position == 0
            assert queue.count == 1
            assert queue.is_empty is False

    def test_queue_add_multiple_captures(self):
        """Test adding multiple captures."""
        from src.capture import CaptureMode
        from src.queue import CaptureQueue

        with patch("src.queue.config.get_setting", return_value=50):
            queue = CaptureQueue()

            for _i in range(5):
                mock_result = MagicMock()
                mock_result.success = True
                queue.add(mock_result, CaptureMode.REGION)

            assert queue.count == 5

    def test_queue_get_all(self):
        """Test get_all returns all results."""
        from src.capture import CaptureMode
        from src.queue import CaptureQueue

        with patch("src.queue.config.get_setting", return_value=50):
            queue = CaptureQueue()

            results = []
            for i in range(3):
                mock_result = MagicMock()
                mock_result.value = i
                results.append(mock_result)
                queue.add(mock_result, CaptureMode.REGION)

            all_results = queue.get_all()
            assert len(all_results) == 3
            for i, result in enumerate(all_results):
                assert result.value == i

    def test_queue_get_at_valid_index(self):
        """Test get_at with valid index."""
        from src.capture import CaptureMode
        from src.queue import CaptureQueue

        with patch("src.queue.config.get_setting", return_value=50):
            queue = CaptureQueue()

            mock_result = MagicMock()
            mock_result.value = 42
            queue.add(mock_result, CaptureMode.REGION)

            result = queue.get_at(0)
            assert result.value == 42

    def test_queue_get_at_invalid_index(self):
        """Test get_at with invalid index returns None."""
        from src.queue import CaptureQueue

        with patch("src.queue.config.get_setting", return_value=50):
            queue = CaptureQueue()

            result = queue.get_at(0)
            assert result is None

            result = queue.get_at(-1)
            assert result is None

    def test_queue_clear(self):
        """Test clearing the queue."""
        from src.capture import CaptureMode
        from src.queue import CaptureQueue

        with patch("src.queue.config.get_setting", return_value=50):
            queue = CaptureQueue()

            for _i in range(3):
                queue.add(MagicMock(), CaptureMode.REGION)

            assert queue.count == 3

            queue.clear()

            assert queue.count == 0
            assert queue.is_empty is True

    def test_queue_pop_all(self):
        """Test pop_all returns all and clears queue."""
        from src.capture import CaptureMode
        from src.queue import CaptureQueue

        with patch("src.queue.config.get_setting", return_value=50):
            queue = CaptureQueue()

            for _i in range(3):
                queue.add(MagicMock(), CaptureMode.REGION)

            results = queue.pop_all()

            assert len(results) == 3
            assert queue.count == 0

    def test_queue_remove_valid_index(self):
        """Test remove with valid index."""
        from src.capture import CaptureMode
        from src.queue import CaptureQueue

        with patch("src.queue.config.get_setting", return_value=50):
            queue = CaptureQueue()

            for _i in range(3):
                queue.add(MagicMock(), CaptureMode.REGION)

            result = queue.remove(1)

            assert result is True
            assert queue.count == 2

    def test_queue_remove_invalid_index(self):
        """Test remove with invalid index."""
        from src.queue import CaptureQueue

        with patch("src.queue.config.get_setting", return_value=50):
            queue = CaptureQueue()

            result = queue.remove(0)
            assert result is False

            result = queue.remove(-1)
            assert result is False

    def test_queue_enforces_max_size(self):
        """Test queue enforces max size limit."""
        from src.capture import CaptureMode
        from src.queue import CaptureQueue

        with patch("src.queue.config.get_setting", return_value=5):
            queue = CaptureQueue()

            for _i in range(10):
                queue.add(MagicMock(), CaptureMode.REGION)

            # Should only keep max_queue_size items
            assert queue.count <= 5


class TestCaptureQueueWithPersistence:
    """Test CaptureQueue with persistence directory."""

    def test_queue_with_persist_dir_init(self):
        """Test queue initialization with persist dir."""
        from src.queue import CaptureQueue

        with tempfile.TemporaryDirectory() as tmpdir:
            persist_dir = Path(tmpdir) / "queue"

            with patch("src.queue.config.get_setting", return_value=50):
                queue = CaptureQueue(persist_dir=persist_dir)

                assert queue.count == 0

    def test_persist_dir_stores_path(self):
        """Test persist_dir is stored."""
        from src.queue import CaptureQueue

        with tempfile.TemporaryDirectory() as tmpdir:
            persist_dir = Path(tmpdir) / "queue"

            with patch("src.queue.config.get_setting", return_value=50):
                queue = CaptureQueue(persist_dir=persist_dir)

                assert queue._persist_dir == persist_dir

    def test_add_with_persist_calls_persist_capture(self):
        """Test add with persistence calls _persist_capture."""
        from src.capture import CaptureMode
        from src.queue import CaptureQueue

        with tempfile.TemporaryDirectory() as tmpdir:
            persist_dir = Path(tmpdir) / "queue"
            persist_dir.mkdir()

            with patch("src.queue.config.get_setting", return_value=50):
                queue = CaptureQueue(persist_dir=persist_dir)

                mock_result = MagicMock()
                mock_result.success = True
                mock_pixbuf = MagicMock()
                mock_result.pixbuf = mock_pixbuf

                queue.add(mock_result, CaptureMode.REGION)

                # Verify pixbuf.savev was called
                mock_pixbuf.savev.assert_called_once()

    def test_clear_with_persist_removes_files(self):
        """Test clear with persistence removes temp files."""
        from src.capture import CaptureMode
        from src.queue import CaptureQueue, QueuedCapture

        with tempfile.TemporaryDirectory() as tmpdir:
            persist_dir = Path(tmpdir) / "queue"
            persist_dir.mkdir()

            # Create a temp file
            temp_file = persist_dir / "test.png"
            temp_file.touch()

            with patch("src.queue.config.get_setting", return_value=50):
                queue = CaptureQueue(persist_dir=persist_dir)

                # Manually add a queued capture with temp_path
                queued = QueuedCapture(
                    result=MagicMock(),
                    timestamp=datetime.now(),
                    mode=CaptureMode.REGION,
                    temp_path=temp_file,
                )
                queue._queue.append(queued)

                assert temp_file.exists()

                queue.clear()

                # File should be deleted
                assert not temp_file.exists()

    def test_remove_with_persist_removes_file(self):
        """Test remove with persistence removes temp file."""
        from src.capture import CaptureMode
        from src.queue import CaptureQueue, QueuedCapture

        with tempfile.TemporaryDirectory() as tmpdir:
            persist_dir = Path(tmpdir) / "queue"
            persist_dir.mkdir()

            temp_file = persist_dir / "test.png"
            temp_file.touch()

            with patch("src.queue.config.get_setting", return_value=50):
                queue = CaptureQueue(persist_dir=persist_dir)

                queued = QueuedCapture(
                    result=MagicMock(),
                    timestamp=datetime.now(),
                    mode=CaptureMode.REGION,
                    temp_path=temp_file,
                )
                queue._queue.append(queued)

                assert temp_file.exists()

                queue.remove(0)

                assert not temp_file.exists()


class TestCaptureQueuePersistCapture:
    """Test CaptureQueue._persist_capture method."""

    def test_persist_capture_raises_without_dir(self):
        """Test _persist_capture raises error without persist_dir."""
        from src.capture import CaptureMode
        from src.queue import CaptureQueue, QueuedCapture

        with patch("src.queue.config.get_setting", return_value=50):
            queue = CaptureQueue()  # No persist_dir

            queued = QueuedCapture(
                result=MagicMock(),
                timestamp=datetime.now(),
                mode=CaptureMode.REGION,
            )

            with pytest.raises(ValueError, match="Persist directory not set"):
                queue._persist_capture(queued)

    def test_persist_capture_creates_directory(self):
        """Test _persist_capture creates persist directory if needed."""
        from src.capture import CaptureMode
        from src.queue import CaptureQueue, QueuedCapture

        with tempfile.TemporaryDirectory() as tmpdir:
            persist_dir = Path(tmpdir) / "new_queue_dir"

            with patch("src.queue.config.get_setting", return_value=50):
                queue = CaptureQueue(persist_dir=persist_dir)

                mock_result = MagicMock()
                mock_pixbuf = MagicMock()
                mock_result.pixbuf = mock_pixbuf

                queued = QueuedCapture(
                    result=mock_result,
                    timestamp=datetime.now(),
                    mode=CaptureMode.REGION,
                )

                queue._persist_capture(queued)

                assert persist_dir.exists()

    def test_persist_capture_handles_savev_exception(self):
        """Test _persist_capture handles pixbuf.savev exception."""
        from src.capture import CaptureMode
        from src.queue import CaptureQueue, QueuedCapture

        with tempfile.TemporaryDirectory() as tmpdir:
            persist_dir = Path(tmpdir) / "queue"
            persist_dir.mkdir()

            with patch("src.queue.config.get_setting", return_value=50):
                queue = CaptureQueue(persist_dir=persist_dir)

                mock_result = MagicMock()
                mock_pixbuf = MagicMock()
                mock_pixbuf.savev.side_effect = Exception("Save failed")
                mock_result.pixbuf = mock_pixbuf

                queued = QueuedCapture(
                    result=mock_result,
                    timestamp=datetime.now(),
                    mode=CaptureMode.REGION,
                )

                # Should return None on failure
                result = queue._persist_capture(queued)
                assert result is None


class TestCaptureQueueLoadPersisted:
    """Test CaptureQueue._load_persisted method."""

    def test_load_persisted_skips_if_no_dir(self):
        """Test _load_persisted does nothing without persist_dir."""
        from src.queue import CaptureQueue

        with patch("src.queue.config.get_setting", return_value=50):
            queue = CaptureQueue()

            # Should not raise
            queue._load_persisted()
            assert queue.count == 0

    def test_load_persisted_skips_if_dir_not_exists(self):
        """Test _load_persisted does nothing if dir doesn't exist."""
        from src.queue import CaptureQueue

        with patch("src.queue.config.get_setting", return_value=50):
            queue = CaptureQueue(persist_dir=Path("/nonexistent/path"))

            # Should not raise
            queue._load_persisted()
            assert queue.count == 0

    def test_load_persisted_skips_without_gtk(self):
        """Test _load_persisted skips if GTK not available."""
        from src.queue import CaptureQueue

        with tempfile.TemporaryDirectory() as tmpdir:
            persist_dir = Path(tmpdir) / "queue"
            persist_dir.mkdir()

            # Create a mock queue file
            queue_file = persist_dir / "queue_20240615_120000_000000.png"
            queue_file.touch()

            with patch("src.queue.config.get_setting", return_value=50):
                with patch("src.queue.GTK_AVAILABLE", False):
                    queue = CaptureQueue(persist_dir=persist_dir)

                    # Should not load anything without GTK
                    assert queue.count == 0


class TestCaptureQueueEdgeCases:
    """Test CaptureQueue edge cases."""

    def test_get_at_negative_index(self):
        """Test get_at with negative index returns None."""
        from src.queue import CaptureQueue

        with patch("src.queue.config.get_setting", return_value=50):
            queue = CaptureQueue()
            queue.add(MagicMock())

            # Python negative indexing should not work here
            result = queue.get_at(-1)
            assert result is None

    def test_get_at_out_of_bounds(self):
        """Test get_at with out of bounds index."""
        from src.queue import CaptureQueue

        with patch("src.queue.config.get_setting", return_value=50):
            queue = CaptureQueue()
            queue.add(MagicMock())

            result = queue.get_at(100)
            assert result is None

    def test_remove_handles_oserror_on_unlink(self):
        """Test remove handles OSError when deleting temp file."""
        from src.capture import CaptureMode
        from src.queue import CaptureQueue, QueuedCapture

        with tempfile.TemporaryDirectory() as tmpdir:
            persist_dir = Path(tmpdir) / "queue"
            persist_dir.mkdir()

            with patch("src.queue.config.get_setting", return_value=50):
                queue = CaptureQueue(persist_dir=persist_dir)

                # Create a mock path that raises OSError on unlink
                mock_path = MagicMock()
                mock_path.exists.return_value = True
                mock_path.unlink.side_effect = OSError("Permission denied")

                queued = QueuedCapture(
                    result=MagicMock(),
                    timestamp=datetime.now(),
                    mode=CaptureMode.REGION,
                    temp_path=mock_path,
                )
                queue._queue.append(queued)

                # Should not raise
                result = queue.remove(0)
                assert result is True
                assert queue.count == 0

    def test_queue_max_size_from_config(self):
        """Test queue uses max size from config."""
        from src.queue import CaptureQueue

        with patch("src.queue.config.get_setting", return_value=10):
            queue = CaptureQueue()
            assert queue._max_queue_size == 10

    def test_add_returns_correct_position(self):
        """Test add returns correct queue position."""
        from src.queue import CaptureQueue

        with patch("src.queue.config.get_setting", return_value=50):
            queue = CaptureQueue()

            pos0 = queue.add(MagicMock())
            pos1 = queue.add(MagicMock())
            pos2 = queue.add(MagicMock())

            assert pos0 == 0
            assert pos1 == 1
            assert pos2 == 2



"""Save/export handler for LikX editor."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Callable, List, Optional

import gi

gi.require_version("Gtk", "3.0")
gi.require_version("Gdk", "3.0")
from gi.repository import Gdk, GdkPixbuf, Gtk  # noqa: E402

from .. import config  # noqa: E402
from ..editor import render_elements  # noqa: E402
from ..notification import (  # noqa: E402
    show_screenshot_copied,
    show_screenshot_saved,
    show_upload_error,
    show_upload_success,
)

if TYPE_CHECKING:
    from ..editor import EditorElement
    from ..uploader import Uploader


class SaveHandler:
    """Handles save, upload, and clipboard operations for the editor.

    Uses callback-based design to avoid direct coupling to EditorWindow.
    """

    def __init__(
        self,
        get_pixbuf: Callable[[], GdkPixbuf.Pixbuf],
        get_elements: Callable[[], List[EditorElement]],
        get_parent_window: Callable[[], Optional[Gtk.Window]],
        set_status: Callable[[str], None],
        get_uploader: Callable[[], Uploader],
    ):
        """Initialize the save handler.

        Args:
            get_pixbuf: Callback to get the source pixbuf
            get_elements: Callback to get the list of annotation elements
            get_parent_window: Callback to get the parent window for dialogs
            set_status: Callback to update the status bar
            get_uploader: Callback to get the uploader instance
        """
        self._get_pixbuf = get_pixbuf
        self._get_elements = get_elements
        self._get_parent_window = get_parent_window
        self._set_status = set_status
        self._get_uploader = get_uploader

    def save_with_annotations(self, filepath: Path) -> bool:
        """Save the image with annotations rendered.

        Args:
            filepath: Path to save the image to

        Returns:
            True if save was successful, False otherwise
        """
        try:
            import cairo

            pixbuf = self._get_pixbuf()
            width = pixbuf.get_width()
            height = pixbuf.get_height()

            surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
            ctx = cairo.Context(surface)

            # Draw original image
            Gdk.cairo_set_source_pixbuf(ctx, pixbuf, 0, 0)
            ctx.paint()

            # Render annotations
            elements = self._get_elements()
            if elements:
                render_elements(surface, elements, pixbuf)

            # Convert to pixbuf and save
            data = surface.get_data()
            new_pixbuf = GdkPixbuf.Pixbuf.new_from_data(
                data,
                GdkPixbuf.Colorspace.RGB,
                True,
                8,
                width,
                height,
                cairo.ImageSurface.format_stride_for_width(cairo.FORMAT_ARGB32, width),
            )

            # Determine format
            format_str = filepath.suffix.lstrip(".").lower()
            format_map = {
                "jpg": "jpeg",
                "jpeg": "jpeg",
                "png": "png",
                "bmp": "bmp",
                "gif": "gif",
            }
            pixbuf_format = format_map.get(format_str, "png")

            filepath.parent.mkdir(parents=True, exist_ok=True)
            new_pixbuf.savev(str(filepath), pixbuf_format, [], [])

            return True
        except Exception as e:
            print(f"Save error: {e}")
            return False

    def save(self) -> None:
        """Save the edited screenshot with a file dialog."""
        parent = self._get_parent_window()
        dialog = Gtk.FileChooserDialog(
            title="Save Screenshot",
            parent=parent,
            action=Gtk.FileChooserAction.SAVE,
        )
        dialog.add_buttons(
            Gtk.STOCK_CANCEL,
            Gtk.ResponseType.CANCEL,
            Gtk.STOCK_SAVE,
            Gtk.ResponseType.OK,
        )
        dialog.set_do_overwrite_confirmation(True)

        # Add filters
        for fmt_name, fmt_ext in [
            ("PNG", "*.png"),
            ("JPEG", "*.jpg"),
            ("BMP", "*.bmp"),
            ("GIF", "*.gif"),
        ]:
            filter_fmt = Gtk.FileFilter()
            filter_fmt.set_name(f"{fmt_name} images")
            filter_fmt.add_pattern(fmt_ext)
            dialog.add_filter(filter_fmt)

        # Set default filename
        default_path = config.get_save_path()
        dialog.set_current_folder(str(default_path.parent))
        dialog.set_current_name(default_path.name)

        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            filepath = Path(dialog.get_filename())
            if self.save_with_annotations(filepath):
                self._set_status(f"Saved to {filepath.name}")
                cfg = config.load_config()
                if cfg.get("show_notification", True):
                    show_screenshot_saved(str(filepath))

        dialog.destroy()

    def upload(self) -> None:
        """Upload the screenshot to cloud service."""
        import os
        import tempfile

        fd, temp_path = tempfile.mkstemp(suffix=".png")
        os.close(fd)
        temp_file = Path(temp_path)

        if not self.save_with_annotations(temp_file):
            show_upload_error("Failed to prepare image for upload")
            return

        self._set_status("Uploading...")

        # Upload
        uploader = self._get_uploader()
        success, url, error = uploader.upload(temp_file)

        # Cleanup
        if temp_file.exists():
            temp_file.unlink()

        if success and url:
            uploader.copy_url_to_clipboard(url)
            self._set_status(f"Uploaded: {url}")
            show_upload_success(url)
        else:
            err_msg = error or "Unknown error"
            self._set_status(f"Upload failed: {err_msg}")
            show_upload_error(err_msg)

    def copy_to_clipboard(self) -> None:
        """Copy the edited screenshot to clipboard."""
        try:
            import cairo

            pixbuf = self._get_pixbuf()
            width = pixbuf.get_width()
            height = pixbuf.get_height()

            surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
            ctx = cairo.Context(surface)

            Gdk.cairo_set_source_pixbuf(ctx, pixbuf, 0, 0)
            ctx.paint()

            elements = self._get_elements()
            if elements:
                render_elements(surface, elements, pixbuf)

            # Convert to pixbuf
            data = surface.get_data()
            new_pixbuf = GdkPixbuf.Pixbuf.new_from_data(
                data,
                GdkPixbuf.Colorspace.RGB,
                True,
                8,
                width,
                height,
                cairo.ImageSurface.format_stride_for_width(cairo.FORMAT_ARGB32, width),
            )

            # Copy to clipboard
            clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
            clipboard.set_image(new_pixbuf)
            clipboard.store()

            self._set_status("Copied to clipboard")
            cfg = config.load_config()
            if cfg.get("show_notification", True):
                show_screenshot_copied()
        except Exception as e:
            self._set_status(f"Copy failed: {e}")



"""Global keyboard shortcuts for LikX."""

import logging
import os
import subprocess
from typing import Callable, Dict, List, Tuple


class HotkeyManager:
    """Manages global keyboard shortcuts."""

    # Base path for LikX custom keybindings
    GNOME_SCHEMA = "org.gnome.settings-daemon.plugins.media-keys"
    GNOME_BASE_PATH = "/org/gnome/settings-daemon/plugins/media-keys/custom-keybindings"

    def __init__(self):
        # key_id -> (callback, command)
        self.hotkeys: Dict[str, Tuple[Callable, str]] = {}
        self.desktop_env = self._detect_desktop_environment()
        self._registered_paths: List[str] = []

    def _detect_desktop_environment(self) -> str:
        """Detect the current desktop environment."""
        desktop = os.environ.get("XDG_CURRENT_DESKTOP", "").lower()

        if "gnome" in desktop:
            return "gnome"
        elif "kde" in desktop or "plasma" in desktop:
            return "kde"
        elif "xfce" in desktop:
            return "xfce"
        elif "mate" in desktop:
            return "mate"
        else:
            return "unknown"

    def register_hotkey(
        self, key_combo: str, callback: Callable, command: str, hotkey_id: str = ""
    ) -> bool:
        """Register a global hotkey.

        Args:
            key_combo: Key combination (e.g., '<Control><Shift>F')
            callback: Function to call when hotkey is pressed
            command: Command to execute
            hotkey_id: Unique identifier for this hotkey (e.g., 'fullscreen', 'region')

        Returns:
            True if registration successful
        """
        # Generate ID from command if not provided
        if not hotkey_id:
            if "--" in command:
                hotkey_id = command.split("--")[-1].split()[0]
            else:
                hotkey_id = "default"

        self.hotkeys[hotkey_id] = (callback, command)

        if self.desktop_env == "gnome":
            return self._register_gnome_hotkey(key_combo, command, hotkey_id)
        elif self.desktop_env == "kde":
            return self._register_kde_hotkey(key_combo, command, hotkey_id)

        return False

    def _register_gnome_hotkey(self, key_combo: str, command: str, hotkey_id: str) -> bool:
        """Register hotkey in GNOME with unique path per hotkey."""
        try:
            # Create unique path for this hotkey
            custom_path = f"{self.GNOME_BASE_PATH}/likx-{hotkey_id}/"

            # Get current custom keybindings
            result = subprocess.run(
                ["gsettings", "get", self.GNOME_SCHEMA, "custom-keybindings"],
                capture_output=True,
                text=True,
            )

            if result.returncode == 0:
                current = result.stdout.strip()
                if custom_path not in current:
                    # Add our custom path
                    if current == "@as []":
                        new_value = f"['{custom_path}']"
                    else:
                        # Parse existing and add
                        new_value = current.rstrip("]") + f", '{custom_path}']"

                    subprocess.run(
                        [
                            "gsettings",
                            "set",
                            self.GNOME_SCHEMA,
                            "custom-keybindings",
                            new_value,
                        ],
                        timeout=5,
                    )

                # Set the binding properties
                binding_schema = f"{self.GNOME_SCHEMA}.custom-keybinding:{custom_path}"
                name = f"LikX {hotkey_id.replace('-', ' ').title()}"

                subprocess.run(["gsettings", "set", binding_schema, "name", name], timeout=5)
                subprocess.run(["gsettings", "set", binding_schema, "command", command], timeout=5)
                subprocess.run(
                    ["gsettings", "set", binding_schema, "binding", key_combo], timeout=5
                )

                if custom_path not in self._registered_paths:
                    self._registered_paths.append(custom_path)

                return True
        except Exception as e:
            logging.warning("Failed to register GNOME hotkey '%s': %s", hotkey_id, e)

        return False

    def _register_kde_hotkey(self, key_combo: str, command: str, hotkey_id: str) -> bool:
        """Register hotkey in KDE Plasma via custom .desktop shortcut files.

        Creates a .desktop file in ~/.local/share/kglobalaccel/ and registers
        the shortcut via kwriteconfig5/kwriteconfig6.
        """
        try:
            from pathlib import Path

            # Convert GTK accelerator to KDE format
            kde_combo = self._gtk_to_kde_shortcut(key_combo)
            if not kde_combo:
                return False

            # Write a .desktop service file for the action
            desktop_dir = Path.home() / ".local" / "share" / "applications"
            desktop_dir.mkdir(parents=True, exist_ok=True)
            desktop_file = desktop_dir / f"likx-{hotkey_id}.desktop"
            desktop_file.write_text(
                f"[Desktop Entry]\n"
                f"Type=Application\n"
                f"Name=LikX {hotkey_id.replace('-', ' ').title()}\n"
                f"Exec={command}\n"
                f"NoDisplay=true\n"
                f"X-KDE-GlobalAccel={kde_combo}\n",
                encoding="utf-8",
            )

            # Try kwriteconfig6 first (Plasma 6), fallback to kwriteconfig5
            config_tool = self._find_kde_config_tool()
            if not config_tool:
                return False

            # Register in kglobalshortcutsrc
            service_name = f"likx-{hotkey_id}.desktop"
            result = subprocess.run(
                [
                    config_tool,
                    "--file",
                    "kglobalshortcutsrc",
                    "--group",
                    service_name,
                    "--key",
                    "_launch",
                    f"{kde_combo},none,LikX {hotkey_id.replace('-', ' ').title()}",
                ],
                capture_output=True,
                timeout=5,
            )

            if result.returncode == 0:
                if hotkey_id not in self._registered_paths:
                    self._registered_paths.append(hotkey_id)
                return True
        except Exception as e:
            logging.warning("Failed to register KDE hotkey '%s': %s", hotkey_id, e)

        return False

    @staticmethod
    def _gtk_to_kde_shortcut(gtk_combo: str) -> str:
        """Convert GTK accelerator format to KDE shortcut format.

        Example: '<Super><Shift>S' -> 'Meta+Shift+S'
        """
        mapping = {
            "<Control>": "Ctrl+",
            "<Ctrl>": "Ctrl+",
            "<Shift>": "Shift+",
            "<Alt>": "Alt+",
            "<Super>": "Meta+",
            "<Meta>": "Meta+",
        }

        result = gtk_combo
        for gtk_mod, kde_mod in mapping.items():
            result = result.replace(gtk_mod, kde_mod)

        # Clean up: remove angle brackets from key name if any remain
        result = result.replace("<", "").replace(">", "")
        return result if result else ""

    @staticmethod
    def _find_kde_config_tool() -> str:
        """Find available KDE config tool (kwriteconfig6 or kwriteconfig5)."""
        for tool in ["kwriteconfig6", "kwriteconfig5"]:
            try:
                subprocess.run([tool, "--help"], capture_output=True, timeout=2)
                return tool
            except (FileNotFoundError, subprocess.TimeoutExpired):
                continue
        return ""

    def update_hotkey(self, hotkey_id: str, key_combo: str) -> bool:
        """Update an existing hotkey binding without re-registering command.

        Args:
            hotkey_id: The hotkey identifier (e.g., 'fullscreen', 'region')
            key_combo: New key combination

        Returns:
            True if update successful
        """
        if self.desktop_env == "gnome":
            try:
                custom_path = f"{self.GNOME_BASE_PATH}/likx-{hotkey_id}/"
                binding_schema = f"{self.GNOME_SCHEMA}.custom-keybinding:{custom_path}"

                result = subprocess.run(
                    ["gsettings", "set", binding_schema, "binding", key_combo],
                    capture_output=True,
                )
                return result.returncode == 0
            except Exception as e:
                logging.warning("Failed to update hotkey '%s': %s", hotkey_id, e)
                return False
        return False

    def unregister_all(self) -> None:
        """Unregister all LikX hotkeys."""
        if self.desktop_env == "gnome":
            try:
                result = subprocess.run(
                    ["gsettings", "get", self.GNOME_SCHEMA, "custom-keybindings"],
                    capture_output=True,
                    text=True,
                    timeout=5,
                )

                if result.returncode == 0:
                    current = result.stdout.strip()

                    for path in self._registered_paths:
                        current = current.replace(f"'{path}', ", "")
                        current = current.replace(f", '{path}'", "")
                        current = current.replace(f"'{path}'", "")

                    if current in ("[]", "[@as ]", ""):
                        current = "@as []"

                    subprocess.run(
                        [
                            "gsettings",
                            "set",
                            self.GNOME_SCHEMA,
                            "custom-keybindings",
                            current,
                        ],
                        timeout=5,
                    )

            except Exception as e:
                logging.warning("Failed to unregister GNOME hotkeys: %s", e)

        elif self.desktop_env == "kde":
            try:
                from pathlib import Path

                desktop_dir = Path.home() / ".local" / "share" / "applications"
                config_tool = self._find_kde_config_tool()

                for hotkey_id in self._registered_paths:
                    # Remove .desktop file
                    desktop_file = desktop_dir / f"likx-{hotkey_id}.desktop"
                    if desktop_file.exists():
                        desktop_file.unlink()

                    # Remove from kglobalshortcutsrc
                    if config_tool:
                        subprocess.run(
                            [
                                config_tool,
                                "--file",
                                "kglobalshortcutsrc",
                                "--group",
                                f"likx-{hotkey_id}.desktop",
                                "--delete",
                                "_launch",
                            ],
                            capture_output=True,
                            timeout=5,
                        )
            except Exception as e:
                logging.warning("Failed to unregister KDE hotkeys: %s", e)

        self._registered_paths.clear()
        self.hotkeys.clear()



"""Command palette UI for LikX - searchable command interface."""

from __future__ import annotations

try:
    import gi

    gi.require_version("Gtk", "3.0")
    gi.require_version("Gdk", "3.0")
    from gi.repository import Gdk, Gtk

    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

from typing import List, Optional

from .commands import Command


class CommandPalette(Gtk.Window):
    """Searchable command palette overlay."""

    def __init__(self, commands: List[Command], parent: Optional[Gtk.Window] = None):
        if not GTK_AVAILABLE:
            raise RuntimeError("GTK not available")

        super().__init__(type=Gtk.WindowType.POPUP)

        self.commands = commands
        self.filtered_commands = commands.copy()
        self.selected_index = 0
        self.parent_window = parent

        self._setup_window()
        self._build_ui()
        self._apply_css()
        self._connect_signals()

    def _setup_window(self):
        """Configure window properties."""
        self.set_decorated(False)
        self.set_modal(True)
        self.set_type_hint(Gdk.WindowTypeHint.POPUP_MENU)
        self.set_skip_taskbar_hint(True)
        self.set_skip_pager_hint(True)
        self.set_keep_above(True)
        self.set_default_size(500, 400)

        if self.parent_window:
            self.set_transient_for(self.parent_window)

    def _build_ui(self):
        """Build the palette UI."""
        # Main container with border
        frame = Gtk.Frame()
        frame.set_shadow_type(Gtk.ShadowType.OUT)
        self.add(frame)

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)
        frame.add(vbox)

        # Search entry with icon
        search_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=8)
        search_box.set_margin_top(12)
        search_box.set_margin_bottom(8)
        search_box.set_margin_start(12)
        search_box.set_margin_end(12)

        prompt_label = Gtk.Label(label=">")
        prompt_label.get_style_context().add_class("prompt")
        search_box.pack_start(prompt_label, False, False, 0)

        self.search_entry = Gtk.Entry()
        self.search_entry.set_placeholder_text("Type to search commands...")
        self.search_entry.set_has_frame(False)
        search_box.pack_start(self.search_entry, True, True, 0)

        vbox.pack_start(search_box, False, False, 0)

        # Separator
        separator = Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL)
        vbox.pack_start(separator, False, False, 0)

        # Results list in scrolled window
        scrolled = Gtk.ScrolledWindow()
        scrolled.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        scrolled.set_min_content_height(300)

        self.listbox = Gtk.ListBox()
        self.listbox.set_selection_mode(Gtk.SelectionMode.SINGLE)
        scrolled.add(self.listbox)

        vbox.pack_start(scrolled, True, True, 0)

        # Footer with hints
        footer = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=12)
        footer.set_margin_top(8)
        footer.set_margin_bottom(8)
        footer.set_margin_start(12)
        footer.set_margin_end(12)

        hints = [("↑↓", "Navigate"), ("Enter", "Execute"), ("Esc", "Close")]
        for key, action in hints:
            hint_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=4)
            key_label = Gtk.Label(label=key)
            key_label.get_style_context().add_class("keyhint")
            action_label = Gtk.Label(label=action)
            action_label.get_style_context().add_class("hint")
            hint_box.pack_start(key_label, False, False, 0)
            hint_box.pack_start(action_label, False, False, 0)
            footer.pack_start(hint_box, False, False, 0)

        vbox.pack_start(footer, False, False, 0)

        # Populate initial list
        self._populate_list()

    def _apply_css(self):
        """Apply dark theme CSS styling."""
        css = b"""
        window {
            background-color: #1e1e1e;
            border: 1px solid #3c3c3c;
            border-radius: 8px;
        }
        frame {
            background-color: #1e1e1e;
            border-radius: 8px;
        }
        entry {
            background-color: transparent;
            color: #ffffff;
            font-size: 16px;
            border: none;
            box-shadow: none;
        }
        entry:focus {
            box-shadow: none;
            border: none;
        }
        .prompt {
            color: #569cd6;
            font-size: 18px;
            font-weight: bold;
        }
        separator {
            background-color: #3c3c3c;
            min-height: 1px;
        }
        list {
            background-color: #1e1e1e;
        }
        row {
            padding: 8px 12px;
            background-color: #1e1e1e;
        }
        row:selected {
            background-color: #094771;
        }
        row:hover {
            background-color: #2d2d2d;
        }
        .command-icon {
            font-size: 16px;
            min-width: 24px;
        }
        .command-name {
            color: #ffffff;
            font-size: 14px;
        }
        .command-shortcut {
            color: #808080;
            font-size: 12px;
        }
        .keyhint {
            background-color: #3c3c3c;
            color: #cccccc;
            padding: 2px 6px;
            border-radius: 3px;
            font-size: 11px;
        }
        .hint {
            color: #808080;
            font-size: 11px;
        }
        """
        provider = Gtk.CssProvider()
        provider.load_from_data(css)
        Gtk.StyleContext.add_provider_for_screen(
            Gdk.Screen.get_default(), provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )

    def _connect_signals(self):
        """Connect event handlers."""
        self.search_entry.connect("changed", self._on_search_changed)
        self.search_entry.connect("key-press-event", self._on_key_press)
        self.listbox.connect("row-activated", self._on_row_activated)
        self.connect("focus-out-event", self._on_focus_out)

    def _populate_list(self):
        """Populate listbox with filtered commands."""
        # Clear existing rows
        for child in self.listbox.get_children():
            self.listbox.remove(child)

        # Add filtered commands
        for cmd in self.filtered_commands:
            row = self._create_command_row(cmd)
            self.listbox.add(row)

        self.listbox.show_all()

        # Select first item
        if self.filtered_commands:
            first_row = self.listbox.get_row_at_index(0)
            if first_row:
                self.listbox.select_row(first_row)
            self.selected_index = 0

    def _create_command_row(self, cmd: Command) -> Gtk.ListBoxRow:
        """Create a row widget for a command."""
        row = Gtk.ListBoxRow()
        row.command = cmd  # Store reference

        hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=12)
        hbox.set_margin_top(4)
        hbox.set_margin_bottom(4)

        # Icon
        icon_label = Gtk.Label(label=cmd.icon if cmd.icon else "  ")
        icon_label.get_style_context().add_class("command-icon")
        hbox.pack_start(icon_label, False, False, 0)

        # Name
        name_label = Gtk.Label(label=cmd.name)
        name_label.get_style_context().add_class("command-name")
        name_label.set_xalign(0)
        hbox.pack_start(name_label, True, True, 0)

        # Shortcut hint
        if cmd.shortcut:
            shortcut_label = Gtk.Label(label=cmd.shortcut)
            shortcut_label.get_style_context().add_class("command-shortcut")
            hbox.pack_end(shortcut_label, False, False, 0)

        row.add(hbox)
        return row

    def _on_search_changed(self, entry):
        """Filter commands based on search text."""
        query = entry.get_text()

        self.filtered_commands = [cmd for cmd in self.commands if cmd.matches(query)]
        self._populate_list()

    def _on_key_press(self, widget, event):
        """Handle keyboard navigation."""
        keyval = event.keyval

        if keyval == Gdk.KEY_Escape:
            self.hide()
            return True

        elif keyval == Gdk.KEY_Return or keyval == Gdk.KEY_KP_Enter:
            self._execute_selected()
            return True

        elif keyval == Gdk.KEY_Up:
            self._move_selection(-1)
            return True

        elif keyval == Gdk.KEY_Down:
            self._move_selection(1)
            return True

        elif keyval == Gdk.KEY_Tab:
            # Tab also moves down
            self._move_selection(1)
            return True

        return False

    def _move_selection(self, delta: int):
        """Move selection up or down."""
        if not self.filtered_commands:
            return

        self.selected_index = (self.selected_index + delta) % len(self.filtered_commands)
        row = self.listbox.get_row_at_index(self.selected_index)
        if row:
            self.listbox.select_row(row)
            # Scroll to make visible
            row.grab_focus()
            self.search_entry.grab_focus()  # Keep focus on entry

    def _execute_selected(self):
        """Execute the currently selected command."""
        row = self.listbox.get_selected_row()
        if row and hasattr(row, "command"):
            cmd = row.command
            self.hide()
            if cmd.callback:
                cmd.callback()

    def _on_row_activated(self, listbox, row):
        """Handle row click/activation."""
        if row and hasattr(row, "command"):
            cmd = row.command
            self.hide()
            if cmd.callback:
                cmd.callback()

    def _on_focus_out(self, widget, event):
        """Hide when focus is lost."""
        # Small delay to allow click handling
        from gi.repository import GLib

        GLib.timeout_add(100, self._check_focus)
        return False

    def _check_focus(self):
        """Check if we should hide due to focus loss."""
        if not self.has_toplevel_focus():
            self.hide()
        return False

    def show_centered(self, parent: Gtk.Window):
        """Show the palette centered on parent window."""
        self.parent_window = parent
        self.set_transient_for(parent)

        # Get parent geometry
        parent_win = parent.get_window()
        if parent_win:
            _, px, py = parent_win.get_origin()
            pw = parent.get_allocated_width()
            ph = parent.get_allocated_height()

            # Calculate centered position
            w, h = 500, 400
            x = px + (pw - w) // 2
            y = py + (ph - h) // 3  # Slightly above center

            self.move(x, y)

        # Reset state
        self.search_entry.set_text("")
        self.filtered_commands = self.commands.copy()
        self._populate_list()

        self.show_all()
        self.search_entry.grab_focus()



"""Cloud upload functionality for LikX."""

import json
import subprocess
from pathlib import Path
from typing import Optional, Tuple

from . import config


def _get_secret(key: str) -> str:
    """Retrieve a secret from system keyring, falling back to config.

    Uses the Secret Service API (GNOME Keyring / KDE Wallet) when available.
    Falls back to config.json for backwards compatibility.
    """
    try:
        import keyring

        value = keyring.get_password("likx", key)
        if value:
            return value
    except ImportError:
        pass

    # Fallback to config (legacy / no keyring installed)
    cfg = config.load_config()
    return cfg.get(key, "")


def _set_secret(key: str, value: str) -> bool:
    """Store a secret in system keyring.

    Returns True if stored in keyring, False if falling back to config.
    """
    try:
        import keyring

        keyring.set_password("likx", key, value)
        return True
    except ImportError:
        # No keyring available — store in config (with warning)
        cfg = config.load_config()
        cfg[key] = value
        config.save_config(cfg)
        return False


class Uploader:
    """Handles uploading screenshots to cloud services."""

    def __init__(self):
        import os

        cfg = config.load_config()
        self.imgur_client_id = (
            os.environ.get("LIKX_IMGUR_CLIENT_ID")
            or cfg.get("imgur_client_id")
            or "546c25a59c58ad7"
        )

    def upload(self, filepath: Path) -> Tuple[bool, Optional[str], Optional[str]]:
        """Upload image using configured service.

        Args:
            filepath: Path to image file

        Returns:
            Tuple of (success, url, error_message)
        """
        cfg = config.load_config()
        service = cfg.get("upload_service", "imgur")

        if service == "imgur":
            return self.upload_to_imgur(filepath)
        elif service == "fileio":
            return self.upload_to_file_io(filepath)
        elif service == "s3":
            return self.upload_to_s3(filepath)
        elif service == "dropbox":
            return self.upload_to_dropbox(filepath)
        elif service == "gdrive":
            return self.upload_to_gdrive(filepath)
        elif service == "none":
            return False, None, "Upload disabled"
        else:
            return False, None, f"Unknown upload service: {service}"

    def _http_upload(
        self,
        url: str,
        *,
        method: str = "POST",
        headers: Optional[dict] = None,
        data: Optional[bytes] = None,
        files: Optional[dict] = None,
        json_body: Optional[dict] = None,
        timeout: int = 30,
    ) -> Tuple[int, dict]:
        """HTTP upload using urllib (stdlib). Returns (status_code, response_json).

        Falls back to curl only if absolutely necessary.
        """
        import urllib.error
        import urllib.request

        req_headers = headers or {}

        if json_body is not None:
            body = json.dumps(json_body).encode("utf-8")
            req_headers.setdefault("Content-Type", "application/json")
        elif files is not None:
            # Multipart form data
            boundary = "----LikXBoundary"
            parts = []
            for field_name, (filename, file_data, content_type) in files.items():
                parts.append(f"--{boundary}\r\n".encode())
                parts.append(
                    f'Content-Disposition: form-data; name="{field_name}"; '
                    f'filename="{filename}"\r\n'.encode()
                )
                parts.append(f"Content-Type: {content_type}\r\n\r\n".encode())
                parts.append(file_data)
                parts.append(b"\r\n")
            parts.append(f"--{boundary}--\r\n".encode())
            body = b"".join(parts)
            req_headers["Content-Type"] = f"multipart/form-data; boundary={boundary}"
        elif data is not None:
            body = data
        else:
            body = None

        req = urllib.request.Request(url, data=body, headers=req_headers, method=method)

        try:
            with urllib.request.urlopen(req, timeout=timeout) as resp:
                response_body = resp.read().decode("utf-8")
                try:
                    return resp.status, json.loads(response_body)
                except json.JSONDecodeError:
                    return resp.status, {"raw": response_body}
        except urllib.error.HTTPError as e:
            response_body = e.read().decode("utf-8", errors="replace")
            try:
                return e.code, json.loads(response_body)
            except json.JSONDecodeError:
                return e.code, {"error": response_body}
        except urllib.error.URLError as e:
            return 0, {"error": str(e.reason)}

    def upload_to_imgur(self, filepath: Path) -> Tuple[bool, Optional[str], Optional[str]]:
        """Upload image to Imgur.

        Args:
            filepath: Path to image file

        Returns:
            Tuple of (success, url, error_message)
        """
        try:
            import base64

            with open(filepath, "rb") as f:
                image_data = base64.b64encode(f.read()).decode("utf-8")

            status, response = self._http_upload(
                "https://api.imgur.com/3/image",
                headers={"Authorization": f"Client-ID {self.imgur_client_id}"},
                data=f"image={image_data}".encode(),
                timeout=30,
            )

            if response.get("success"):
                url = response["data"]["link"]
                return True, url, None
            else:
                return False, None, "Imgur API returned error"

        except FileNotFoundError:
            return False, None, f"Image file not found: {filepath}"
        except OSError as e:
            return False, None, str(e)

    def upload_to_file_io(self, filepath: Path) -> Tuple[bool, Optional[str], Optional[str]]:
        """Upload to file.io (temporary file sharing).

        Args:
            filepath: Path to image file

        Returns:
            Tuple of (success, url, error_message)
        """
        try:
            with open(filepath, "rb") as f:
                file_data = f.read()

            status, response = self._http_upload(
                "https://file.io",
                files={"file": (filepath.name, file_data, "application/octet-stream")},
                timeout=30,
            )

            if response.get("success"):
                url = response.get("link", "")
                return True, url, None
            else:
                return False, None, "file.io returned error"

        except FileNotFoundError:
            return False, None, f"Image file not found: {filepath}"
        except OSError as e:
            return False, None, str(e)

    def upload_to_s3(self, filepath: Path) -> Tuple[bool, Optional[str], Optional[str]]:
        """Upload image to AWS S3.

        Requires AWS CLI configured or environment variables:
        - AWS_ACCESS_KEY_ID
        - AWS_SECRET_ACCESS_KEY
        - S3 bucket configured in LikX settings

        Args:
            filepath: Path to image file

        Returns:
            Tuple of (success, url, error_message)
        """
        cfg = config.load_config()
        bucket = cfg.get("s3_bucket", "")
        region = cfg.get("s3_region", "us-east-1")
        make_public = cfg.get("s3_public", True)

        if not bucket:
            return False, None, "S3 bucket not configured in settings"

        try:
            s3_key = f"screenshots/{filepath.name}"

            cmd = [
                "aws",
                "s3",
                "cp",
                str(filepath),
                f"s3://{bucket}/{s3_key}",
                "--region",
                region,
            ]

            if make_public:
                cmd.extend(["--acl", "public-read"])

            result = subprocess.run(
                cmd,
                capture_output=True,
                text=True,
                timeout=60,
                shell=False,
            )

            if result.returncode == 0:
                if make_public:
                    url = f"https://{bucket}.s3.{region}.amazonaws.com/{s3_key}"
                else:
                    url = f"s3://{bucket}/{s3_key}"
                return True, url, None
            else:
                error = result.stderr.strip() or "Upload failed"
                return False, None, error

        except subprocess.TimeoutExpired:
            return False, None, "Upload timed out"
        except FileNotFoundError:
            return (
                False,
                None,
                "AWS CLI not installed. Install with: pip install awscli",
            )
        except OSError as e:
            return False, None, str(e)

    def _dropbox_upload_file(
        self, filepath: Path, access_token: str, dropbox_path: str
    ) -> Tuple[bool, Optional[str]]:
        """Upload file to Dropbox. Returns (success, error_message)."""
        with open(filepath, "rb") as f:
            file_data = f.read()

        # Use json.dumps for safe serialization (prevents JSON injection)
        api_arg = json.dumps({"path": dropbox_path, "mode": "add"})

        status, response = self._http_upload(
            "https://content.dropboxapi.com/2/files/upload",
            headers={
                "Authorization": f"Bearer {access_token}",
                "Content-Type": "application/octet-stream",
                "Dropbox-API-Arg": api_arg,
            },
            data=file_data,
            timeout=60,
        )

        if "error" in response:
            return False, response.get("error_summary", "Upload failed")
        if status == 0:
            return False, response.get("error", "Upload request failed")
        return True, None

    def _dropbox_create_share_link(self, access_token: str, dropbox_path: str) -> Optional[str]:
        """Create Dropbox shared link. Returns URL or None."""
        status, response = self._http_upload(
            "https://api.dropboxapi.com/2/sharing/create_shared_link_with_settings",
            headers={
                "Authorization": f"Bearer {access_token}",
            },
            json_body={"path": dropbox_path},
            timeout=30,
        )

        if "url" in response:
            return response["url"].replace("dl=0", "dl=1")
        if "shared_link_already_exists" in str(response):
            existing = (
                response.get("error", {})
                .get("shared_link_already_exists", {})
                .get("metadata", {})
                .get("url", "")
            )
            if existing:
                return existing.replace("dl=0", "dl=1")
        return None

    def upload_to_dropbox(self, filepath: Path) -> Tuple[bool, Optional[str], Optional[str]]:
        """Upload image to Dropbox."""
        access_token = _get_secret("dropbox_token")

        if not access_token:
            return (
                False,
                None,
                "Dropbox access token not configured. "
                "Get one from https://www.dropbox.com/developers/apps",
            )

        try:
            dropbox_path = f"/Screenshots/{filepath.name}"
            success, error = self._dropbox_upload_file(filepath, access_token, dropbox_path)
            if not success:
                return False, None, error

            url = self._dropbox_create_share_link(access_token, dropbox_path)
            return True, url or f"Uploaded to Dropbox: {dropbox_path}", None

        except json.JSONDecodeError:
            return False, None, "Invalid response from Dropbox"
        except OSError as e:
            return False, None, str(e)

    def _gdrive_upload_with_gdrive(
        self, filepath: Path, folder_id: str
    ) -> Tuple[bool, Optional[str], Optional[str]]:
        """Upload using gdrive CLI. Returns (success, url, error)."""
        cmd = ["gdrive", "files", "upload", str(filepath)]
        if folder_id:
            cmd.extend(["--parent", folder_id])

        result = subprocess.run(cmd, capture_output=True, text=True, timeout=60, shell=False)

        if result.returncode == 0:
            output = result.stdout
            if "id" in output.lower():
                parts = output.split()
                for i, part in enumerate(parts):
                    if part.lower() == "id" and i + 1 < len(parts):
                        file_id = parts[i + 1].strip()
                        return True, f"https://drive.google.com/file/d/{file_id}/view", None
            return True, "Uploaded to Google Drive", None

        error = result.stderr.strip() or "Upload failed"
        if "not found" in error.lower() or result.returncode == 127:
            raise FileNotFoundError("gdrive not found")
        return False, None, error

    def _gdrive_upload_with_rclone(
        self, filepath: Path, remote: str
    ) -> Tuple[bool, Optional[str], Optional[str]]:
        """Upload using rclone. Returns (success, url, error)."""
        remote_path = f"{remote}:Screenshots/{filepath.name}"
        result = subprocess.run(
            ["rclone", "copyto", str(filepath), remote_path],
            capture_output=True,
            text=True,
            timeout=60,
            shell=False,
        )

        if result.returncode != 0:
            return False, None, result.stderr.strip() or "Upload failed"

        link_result = subprocess.run(
            ["rclone", "link", remote_path],
            capture_output=True,
            text=True,
            timeout=30,
            shell=False,
        )
        if link_result.returncode == 0 and link_result.stdout.strip():
            return True, link_result.stdout.strip(), None
        return True, f"Uploaded to {remote_path}", None

    def upload_to_gdrive(self, filepath: Path) -> Tuple[bool, Optional[str], Optional[str]]:
        """Upload image to Google Drive."""
        cfg = config.load_config()
        folder_id = cfg.get("gdrive_folder_id", "")

        try:
            return self._gdrive_upload_with_gdrive(filepath, folder_id)
        except FileNotFoundError:
            pass  # Try rclone next

        try:
            remote = cfg.get("gdrive_rclone_remote", "gdrive")
            return self._gdrive_upload_with_rclone(filepath, remote)
        except FileNotFoundError:
            return (
                False,
                None,
                "Google Drive upload requires gdrive or rclone. "
                "Install gdrive: https://github.com/glotlabs/gdrive",
            )
        except subprocess.TimeoutExpired:
            return False, None, "Upload timed out"
        except OSError as e:
            return False, None, str(e)

    def copy_url_to_clipboard(self, url: str) -> bool:
        """Copy URL to clipboard.

        Args:
            url: URL to copy

        Returns:
            True if successful
        """
        try:
            subprocess.run(
                ["xclip", "-selection", "clipboard"],
                input=url.encode(),
                check=True,
                shell=False,
            )
            return True
        except (FileNotFoundError, subprocess.CalledProcessError):
            pass

        try:
            subprocess.run(
                ["xsel", "--clipboard", "--input"],
                input=url.encode(),
                check=True,
                shell=False,
            )
            return True
        except (FileNotFoundError, subprocess.CalledProcessError):
            pass

        return False
